from collections.abc import Iterator, Sequence
from typing import Any

class ParameterGrid:
    def __init__(self, *parameters: Sequence): ...
    def __getitem__(self, index: Sequence[int] | int) -> Sequence[Any]: ...
    def __iter__(self) -> Iterator[Sequence[Any]]: ...
    def __len__(self) -> int: ...
    @property
    def size(self) -> int: ...
    def shuffled(
        self, rounds: int = ..., seed: int | None = ...
    ) -> Iterator[Sequence[Any]]: ...
